"use strict";
// How HMR works
// 1. `.vue` files are transformed into `.js` files before being served
// 2. All `.js` files, before being served, are parsed to detect their imports
//    (this is done in `./modules.ts`) for module import rewriting. During this
//    we also record the importer/importee relationships which can be used for
//    HMR analysis (we do both at the same time to avoid double parse costs)
// 3. When a `.vue` file changes, we directly read, parse it again and
//    notify the client because Vue components are self-accepting by nature
// 4. When a js file changes, it triggers an HMR graph analysis, where we try to
//    walk its importer chains and see if we reach a "HMR boundary". An HMR
//    boundary is either a `.vue` file or a `.js` file that explicitly indicated
//    that it accepts hot updates (by importing from the `/@hmr` special module)
// 5. If any parent chain exhausts without ever running into an HMR boundary,
//    it's considered a "dead end". This causes a full page reload.
// 6. If a `.vue` boundary is encountered, we add it to the `vueImports` Set.
// 7. If a `.js` boundary is encountered, we check if the boundary's current
//    child importer is in the accepted list of the boundary (see additional
//    explanation below). If yes, record current child importer in the
//    `jsImporters` Set.
// 8. If the graph walk finished without running into dead ends, notify the
//    client to update all `jsImporters` and `vueImporters`.
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const ws_1 = __importDefault(require("ws"));
const hash_sum_1 = __importDefault(require("hash-sum"));
const serverPluginVue_1 = require("./serverPluginVue");
const utils_1 = require("./utils");
const serverPluginModules_1 = require("./serverPluginModules");
const chalk_1 = __importDefault(require("chalk"));
const debug = require('debug')('vite:hmr');
// client and node files are placed flat in the dist folder
exports.hmrClientFilePath = path_1.default.resolve(__dirname, './client.js');
exports.hmrClientPublicPath = '/@hmr';
exports.hmrPlugin = ({ root, app, server, watcher }) => {
    app.use(async (ctx, next) => {
        if (ctx.path !== exports.hmrClientPublicPath) {
            return next();
        }
        debug('serving hmr client');
        ctx.type = 'js';
        ctx.body = await utils_1.cachedRead(exports.hmrClientFilePath);
    });
    // start a websocket server to send hmr notifications to the client
    const wss = new ws_1.default.Server({ server });
    const sockets = new Set();
    wss.on('connection', (socket) => {
        debug('ws client connected');
        sockets.add(socket);
        socket.send(JSON.stringify({ type: 'connected' }));
        socket.on('close', () => {
            sockets.delete(socket);
        });
    });
    wss.on('error', (e) => {
        if (e.code !== 'EADDRINUSE') {
            console.error(chalk_1.default.red(`[vite] WebSocket server error:`));
            console.error(e);
        }
    });
    const notify = (payload) => {
        const stringified = JSON.stringify(payload, null, 2);
        debug(`update: ${stringified}`);
        sockets.forEach((s) => s.send(stringified));
    };
    watcher.on('change', async (file) => {
        const timestamp = Date.now();
        const servedPath = '/' + path_1.default.relative(root, file);
        if (file.endsWith('.vue')) {
            handleVueSFCReload(file, servedPath, timestamp);
        }
        else {
            handleJSReload(servedPath, timestamp);
        }
    });
    function handleJSReload(servedPath, timestamp) {
        // normal js file
        const importers = serverPluginModules_1.importerMap.get(servedPath);
        if (importers) {
            const vueImporters = new Set();
            const jsHotImporters = new Set();
            const hasDeadEnd = walkImportChain(servedPath, importers, vueImporters, jsHotImporters);
            if (hasDeadEnd) {
                notify({
                    type: 'full-reload',
                    timestamp
                });
            }
            else {
                vueImporters.forEach((vueImporter) => {
                    notify({
                        type: 'vue-reload',
                        path: vueImporter,
                        timestamp
                    });
                });
                jsHotImporters.forEach((jsImporter) => {
                    notify({
                        type: 'js-update',
                        path: jsImporter,
                        timestamp
                    });
                });
            }
        }
    }
    function walkImportChain(importee, currentImporters, vueImporters, jsHotImporters) {
        let hasDeadEnd = false;
        for (const importer of currentImporters) {
            if (importer.endsWith('.vue')) {
                vueImporters.add(importer);
            }
            else if (isHMRBoundary(importer, importee)) {
                jsHotImporters.add(importer);
            }
            else {
                const parentImpoters = serverPluginModules_1.importerMap.get(importer);
                if (!parentImpoters) {
                    hasDeadEnd = true;
                }
                else {
                    hasDeadEnd = walkImportChain(importer, parentImpoters, vueImporters, jsHotImporters);
                }
            }
        }
        return hasDeadEnd;
    }
    function isHMRBoundary(importer, dep) {
        const deps = serverPluginModules_1.hmrBoundariesMap.get(importer);
        return deps ? deps.has(dep) : false;
    }
    async function handleVueSFCReload(file, servedPath, timestamp) {
        const cacheEntry = serverPluginVue_1.vueCache.get(file);
        serverPluginVue_1.vueCache.del(file);
        const descriptor = await serverPluginVue_1.parseSFC(root, file);
        if (!descriptor) {
            // read failed
            return;
        }
        const prevDescriptor = cacheEntry && cacheEntry.descriptor;
        if (!prevDescriptor) {
            // the file has never been accessed yet
            return;
        }
        // check which part of the file changed
        let needReload = false;
        let needRerender = false;
        if (!isEqual(descriptor.script, prevDescriptor.script)) {
            needReload = true;
        }
        if (!isEqual(descriptor.template, prevDescriptor.template)) {
            needRerender = true;
        }
        const styleId = hash_sum_1.default(servedPath);
        const prevStyles = prevDescriptor.styles || [];
        const nextStyles = descriptor.styles || [];
        if ((!needReload &&
            prevStyles.some((s) => s.scoped) !==
                nextStyles.some((s) => s.scoped)) ||
            // TODO for now we force the component to reload on <style module> change
            // but this should be optimizable to replace the __cssMoudles object
            // on script and only trigger a rerender.
            prevStyles.some((s) => s.module != null) ||
            nextStyles.some((s) => s.module != null)) {
            needReload = true;
        }
        // only need to update styles if not reloading, since reload forces
        // style updates as well.
        if (!needReload) {
            nextStyles.forEach((_, i) => {
                if (!prevStyles[i] || !isEqual(prevStyles[i], nextStyles[i])) {
                    notify({
                        type: 'vue-style-update',
                        path: servedPath,
                        index: i,
                        id: `${styleId}-${i}`,
                        timestamp
                    });
                }
            });
        }
        // stale styles always need to be removed
        prevStyles.slice(nextStyles.length).forEach((_, i) => {
            notify({
                type: 'vue-style-remove',
                path: servedPath,
                id: `${styleId}-${i + nextStyles.length}`,
                timestamp
            });
        });
        if (needReload) {
            notify({
                type: 'vue-reload',
                path: servedPath,
                timestamp
            });
        }
        else if (needRerender) {
            notify({
                type: 'vue-rerender',
                path: servedPath,
                timestamp
            });
        }
    }
};
function isEqual(a, b) {
    if (!a || !b)
        return false;
    if (a.content !== b.content)
        return false;
    const keysA = Object.keys(a.attrs);
    const keysB = Object.keys(b.attrs);
    if (keysA.length !== keysB.length) {
        return false;
    }
    return keysA.every((key) => a.attrs[key] === b.attrs[key]);
}
